/*
 * cassette.c - cassette emulation
 *
 * Copyright (C) 2001 Piotr Fusik
 * Copyright (C) 2001-2010 Atari800 development team (see DOC/CREDITS)
 *
 * This file is part of the Atari800 emulator project which emulates
 * the Atari 400, 800, 800XL, 130XE, and 5200 8-bit computers.
 *
 * Atari800 is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Atari800 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Atari800; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "config.h"
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <a8cas.h>
#include <limits.h>
#include <math.h>

#include "atari.h"
#include "cpu.h"
#include "cassette.h"
#include "cfg.h"
#include "esc.h"
#include "log.h"
#include "memory.h"
#include "sio.h"
#include "util.h"
#include "pia.h"
#include "pokey.h"
#include "antic.h"

#ifdef SYNCHRONIZED_SOUND
#include "syncsound.h"
/* Uncomment to write debug informations on audio buffer under/overflow. */
/*#define DEBUG_AUDIO*/
#endif /* SYNCHRONIZED_SOUND */

static A8CAS_FILE *cassette_file = NULL;

char CASSETTE_filename[FILENAME_MAX];
CASSETTE_status_t CASSETTE_status = CASSETTE_STATUS_NONE;
int CASSETTE_write_protect = FALSE;
int CASSETTE_record = FALSE;
static int cassette_writable = FALSE;
static int cassette_readable = FALSE;
static int cassette_samplerate;

char CASSETTE_description[CASSETTE_DESCRIPTION_MAX] = "";


static int cassette_motor = 0;

int CASSETTE_hold_start_on_reboot = 0;
int CASSETTE_hold_start = 0;
int CASSETTE_press_space = 0;
/* Indicates whether the tape has ended. During saving the value is always 0;
   during loading it is equal to (A8CAS_tell() >= A8CAS_size()). */
static int eof_of_tape = 0;

int CASSETTE_position_in_blocks;

static A8CAS_signal current_sig;
static double signal_length;

static unsigned int prev_cpu_clock;

enum CASSETTE_TURBO_TYPE_t CASSETTE_turbo_type = CASSETTE_TURBO_NONE;
static char const * const turbo_type_cfg_strings[CASSETTE_TURBO_TYPE_SIZE] = {
	"NONE",
	"AST-XC12",
	"BLIZZARD",
	"HARD",
	"MANUAL",
	"MANUAL-REV"
};

enum TURBO_CONTROL_t {
	TURBO_CONTROL_NONE,
	TURBO_CONTROL_MANUAL,
	TURBO_CONTROL_DATA_OUT,
	TURBO_CONTROL_COMMAND,
	TURBO_CONTROL_COMMAND_MOTOR
};
static enum TURBO_CONTROL_t turbo_control;
static int turbo_descend_is_1;
int CASSETTE_turbo_state = 0;
static int file_turbo_state = 0;
int CASSETTE_turbo_tolerance = CASSETTE_DEFAULT_TURBO_TOLERANCE;

#ifdef SYNCHRONIZED_SOUND
int CASSETTE_play_audio = 1;
unsigned int CASSETTE_audio_volume = CASSETTE_DEFAULT_AUDIO_VOLUME;
unsigned int CASSETTE_crosstalk_volume = CASSETTE_DEFAULT_CROSSTALK_VOLUME;

static struct {
	unsigned int samplerate;
	unsigned int n_pokeys;
	unsigned int bit16;
	unsigned int buf_size;
	int write_pos;
	double sample_pos;
	unsigned int samples_left_in_signal;
	UBYTE *buffer; /* When BUFFER is NULL, it indicates that audio was not initialised. */
} audio = { 44100, 2, 16, 0, 0, 0.0, 0, NULL };
#endif /* SYNCHRONIZED_SOUND */

static void CassetteWrite(int num_ticks);
static void CassetteRead(int num_ticks);
static void (*cassette_write_func)(int) = &CassetteWrite;
static void (*cassette_read_func)(int) = &CassetteRead;

/* For operations with SIO-patch */
static int gap_delay;

static void UpdateFlags(void)
{
	cassette_readable = cassette_motor &&
	                    (CASSETTE_status == CASSETTE_STATUS_READ_WRITE ||
	                     CASSETTE_status == CASSETTE_STATUS_READ_ONLY) &&
	                     !eof_of_tape;
	cassette_writable = cassette_motor &&
	                    (CASSETTE_status == CASSETTE_STATUS_READ_WRITE ||
	                     CASSETTE_status == CASSETTE_STATUS_WRITE_ONLY) &&
	                    !CASSETTE_write_protect;
}

static void UpdateTurboState(void)
{
	if (CASSETTE_turbo_state != file_turbo_state &&
	    cassette_file != NULL &&
	    !CASSETTE_record) {
		A8CAS_set_param(cassette_file, A8CAS_PARAM_PWM_DECODE, &CASSETTE_turbo_state);
		file_turbo_state = CASSETTE_turbo_state;
	}
}

static void ResetTurboState(void)
{
	switch (turbo_control) {
	case TURBO_CONTROL_DATA_OUT:
		CASSETTE_turbo_state = !POKEY_serial_data_output;
		break;
	case TURBO_CONTROL_COMMAND:
		CASSETTE_turbo_state = !(PIA_PBCTL & 0x08);
		break;
	case TURBO_CONTROL_COMMAND_MOTOR:
		CASSETTE_turbo_state = !(PIA_PBCTL & 0x08) && cassette_motor;
		break;
	default:
		break;
	}
	UpdateTurboState();
}

static void UpdateTurboTolerance(void)
{
	unsigned short int value = CASSETTE_turbo_tolerance * USHRT_MAX / 10 / CASSETTE_MAX_TURBO_TOLERANCE;
	A8CAS_set_param(cassette_file, A8CAS_PARAM_PWM_TOLERANCE, &value);
}

#ifdef SYNCHRONIZED_SOUND
static void UpdateCrosstalkVolume(void)
{
	float value = (float)CASSETTE_crosstalk_volume / CASSETTE_MAX_CROSSTALK_VOLUME;
	A8CAS_set_param(cassette_file, A8CAS_PARAM_CROSSTALK_VOLUME, &value);
}
#endif

int CASSETTE_ReadConfig(char *string, char *ptr)
{
	if (strcmp(string, "CASSETTE_FILENAME") == 0)
		Util_strlcpy(CASSETTE_filename, ptr, sizeof(CASSETTE_filename));
	else if (strcmp(string, "CASSETTE_LOADED") == 0) {
		int value = Util_sscanbool(ptr);
		if (value == -1)
			return FALSE;
		CASSETTE_status = (value ? CASSETTE_STATUS_READ_WRITE : CASSETTE_STATUS_NONE);
	}
	else if (strcmp(string, "CASSETTE_WRITE_PROTECT") == 0) {
		int value = Util_sscanbool(ptr);
		if (value == -1)
			return FALSE;
		CASSETTE_write_protect = value;
	}
	else if (strcmp(string, "CASSETTE_TURBO_TYPE") == 0) {
		int i = CFG_MatchTextParameter(ptr, turbo_type_cfg_strings, CASSETTE_TURBO_TYPE_SIZE);
		if (i < 0)
			return FALSE;
		CASSETTE_SetTurboType(i);
	}
#ifdef SYNCHRONIZED_SOUND
	else if (strcmp(string, "CASSETTE_PLAY_AUDIO") == 0)
		return (CASSETTE_play_audio = Util_sscanbool(ptr)) != -1;
	else if (strcmp(string, "CASSETTE_AUDIO_VOLUME") == 0)
		return (CASSETTE_audio_volume = Util_sscandec(ptr)) != -1;
	else if (strcmp(string, "CASSETTE_CROSSTALK_VOLUME") == 0) {
		int value = Util_sscandec(ptr);
		if (value == -1)
			return FALSE;
		CASSETTE_SetCrosstalkVolume(value);
	}
#endif /* SYNCHRONIZED_SOUND */
	else return FALSE;
	return TRUE;
}
void CASSETTE_WriteConfig(FILE *fp)
{
	fprintf(fp, "CASSETTE_FILENAME=%s\n", CASSETTE_filename);
	fprintf(fp, "CASSETTE_LOADED=%d\n", CASSETTE_status != CASSETTE_STATUS_NONE);
	fprintf(fp, "CASSETTE_WRITE_PROTECT=%d\n", CASSETTE_write_protect);
	fprintf(fp, "CASSETTE_TURBO_TYPE=%s\n", turbo_type_cfg_strings[CASSETTE_turbo_type]);
#ifdef SYNCHRONIZED_SOUND
	fprintf(fp, "CASSETTE_PLAY_AUDIO=%d\n", CASSETTE_play_audio);
	fprintf(fp, "CASSETTE_AUDIO_VOLUME=%u\n", CASSETTE_audio_volume);
	fprintf(fp, "CASSETTE_CROSSTALK_VOLUME=%u\n", CASSETTE_crosstalk_volume);
#endif /* SYNCHRONIZED_SOUND */
}

int CASSETTE_Initialise(int *argc, char *argv[])
{
	int i;
	int j;
	int protect = FALSE; /* Is write-protect requested in command line? */

	prev_cpu_clock = ANTIC_CPU_CLOCK;

	for (i = j = 1; i < *argc; i++) {
		int i_a = (i + 1 < *argc);		/* is argument available? */
		int a_m = FALSE;			/* error, argument missing! */
		int a_i = FALSE;			/* error, argument invalid! */

		if (strcmp(argv[i], "-tape") == 0) {
			if (i_a) {
				Util_strlcpy(CASSETTE_filename, argv[++i], sizeof(CASSETTE_filename));
				CASSETTE_status = CASSETTE_STATUS_READ_WRITE;
				/* Reset any write-protection read from config file. */
				CASSETTE_write_protect = FALSE;
			}
			else a_m = TRUE;
		}
		else if (strcmp(argv[i], "-boottape") == 0) {
			if (i_a) {
				Util_strlcpy(CASSETTE_filename, argv[++i], sizeof(CASSETTE_filename));
				CASSETTE_status = CASSETTE_STATUS_READ_WRITE;
				/* Reset any write-protection read from config file. */
				CASSETTE_write_protect = FALSE;
				CASSETTE_hold_start = 1;
			}
			else a_m = TRUE;
		}
		else if (strcmp(argv[i], "-tape-readonly") == 0)
			protect = TRUE;
		else if (strcmp(argv[i], "-tape-turbo") == 0) {
			if (i_a) {
				int idx = CFG_MatchTextParameter(argv[++i], turbo_type_cfg_strings, CASSETTE_TURBO_TYPE_SIZE);
				if (idx < 0)
					a_i = TRUE;
				else
					CASSETTE_SetTurboType(idx);
			}
			else a_m = TRUE;
		}
		else {
			if (strcmp(argv[i], "-help") == 0) {
				Log_print("\t-tape <file>       Insert cassette image");
				Log_print("\t-boottape <file>   Insert cassette image and boot it");
				Log_print("\t-tape-readonly     Mark the attached cassette image as read-only");
				Log_print("\t-tape-turbo none|ast-xc12|blizzard|hard|manual|manual-rev");
				Log_print("\t                   Choose tape recorder's turbo modification");
			}
			argv[j++] = argv[i];
		}

		if (a_m) {
			Log_print("Missing argument for '%s'", argv[i]);
			return FALSE;
		} else if (a_i) {
			Log_print("Invalid argument for '%s'", argv[--i]);
			return FALSE;
		}
	}

	*argc = j;

	/* If CASSETTE_status was set in this function or in CASSETTE_ReadConfig(),
	   then tape is to be mounted. */
	if (CASSETTE_status != CASSETTE_STATUS_NONE && CASSETTE_filename[0] != '\0') {
		/* Tape is mounted unprotected by default - overrun it if needed. */
		protect = protect || CASSETTE_write_protect;
		CASSETTE_Insert(CASSETTE_filename);
		if (protect)
			CASSETTE_ToggleWriteProtect();
	}

	return TRUE;
}

void CASSETTE_Exit(void)
{
	CASSETTE_Remove();
#ifdef SYNCHRONIZED_SOUND
	if (audio.buffer != NULL) {
		free(audio.buffer);
	}
#endif
}

int CASSETTE_Insert(const char *filename)
{
	A8CAS_info cas_info;

	CASSETTE_Remove();
	/* Guard against providing CASSETTE_filename as parameter. */
	if (CASSETTE_filename != filename)
		strcpy(CASSETTE_filename, filename);
	eof_of_tape = 0;

	cas_info.format = A8CAS_FORMAT_ANY;
	cas_info.mode = A8CAS_RDWR;
	if ((cassette_file = A8CAS_open(CASSETTE_filename, &cas_info)) == NULL) {
		/* Try opening read-only. */
		cas_info.mode = A8CAS_READ;
		if ((cassette_file = A8CAS_open(CASSETTE_filename, &cas_info)) == NULL) {
			/* Try opening as a raw file. */
			cas_info.mode = A8CAS_RDWR;
			cas_info.format = A8CAS_FORMAT_RAW;
			if ((cassette_file = A8CAS_open(CASSETTE_filename, &cas_info)) == NULL) {
				/* Last attempt, try opening as a raw read-only file. */
				cas_info.mode = A8CAS_READ;
				if ((cassette_file = A8CAS_open(CASSETTE_filename, &cas_info)) == NULL)
					return FALSE;
			}
		}
	}
	CASSETTE_status = (cas_info.mode == A8CAS_RDWR ? CASSETTE_STATUS_READ_WRITE : CASSETTE_STATUS_READ_ONLY);
	cassette_samplerate = cas_info.samplerate;
	signal_length = 0.0;
	if (cas_info.description != NULL) {
		Util_strlcpy(CASSETTE_description, cas_info.description, sizeof(CASSETTE_description));
	}

	CASSETTE_write_protect = FALSE;
	CASSETTE_record = FALSE;
	CASSETTE_position_in_blocks = A8CAS_seek_unit(cassette_file) == A8CAS_SEEK_UNIT_BLOCKS;
	UpdateFlags();

#ifdef SYNCHRONIZED_SOUND
	audio.sample_pos = 0.0;
	audio.samples_left_in_signal = 0;
	A8CAS_set_audio(cassette_file, audio.samplerate, (audio.bit16 ? 16 : 8));
	UpdateCrosstalkVolume();
#endif
	file_turbo_state = 0;
	ResetTurboState();
	UpdateTurboTolerance();

	gap_delay = 0;

	if (cas_info.format == A8CAS_FORMAT_RAW)
		return 2;
	else
		return TRUE;
}

void CASSETTE_Remove(void)
{
	if (cassette_file != NULL) {
		A8CAS_close(cassette_file);
		cassette_file = NULL;
	}
	CASSETTE_description[0] = '\0';
	CASSETTE_status = CASSETTE_STATUS_NONE;
	UpdateFlags();
}

static int CreateBlankImage(A8CAS_format format, char const *filename, char *description)
{
	A8CAS_info cas_info;
	A8CAS_FILE *file;

	cas_info.format = format;
	cas_info.samplerate = 44100;
	cas_info.description = description;
	cas_info.mode = A8CAS_WRRD;
	if ((file = A8CAS_open(filename, &cas_info)) == NULL) {
		return FALSE;
	}
	CASSETTE_Remove(); /* Unmount any previous tape image */
	cassette_file = file;
	Util_strlcpy(CASSETTE_filename, filename, sizeof(CASSETTE_filename));
	if (description != NULL)
		Util_strlcpy(CASSETTE_description, description, sizeof(CASSETTE_description));
	CASSETTE_status = (cas_info.mode == A8CAS_WRRD ? CASSETTE_STATUS_READ_WRITE : CASSETTE_STATUS_WRITE_ONLY);
	cassette_samplerate = cas_info.samplerate;

	signal_length = 0.0;

	eof_of_tape = 0;
	CASSETTE_record = TRUE;
	CASSETTE_write_protect = FALSE;
	CASSETTE_position_in_blocks = A8CAS_seek_unit(cassette_file) == A8CAS_SEEK_UNIT_BLOCKS;
	UpdateFlags();
#ifdef SYNCHRONIZED_SOUND
	audio.sample_pos = 0.0;
	audio.samples_left_in_signal = 0;
	A8CAS_set_audio(cassette_file, audio.samplerate, (audio.bit16 ? 16 : 8));
	UpdateCrosstalkVolume();
#endif
	file_turbo_state = 0;
	ResetTurboState();
	UpdateTurboTolerance();

	gap_delay = 0;

	return TRUE;
}

int CASSETTE_CreateCAS(char const *filename, char *description)
{
	return CreateBlankImage(A8CAS_FORMAT_CAS, filename, description);
}

int CASSETTE_CreateWAV(char const *filename)
{
	return CreateBlankImage(A8CAS_FORMAT_SNDFILE, filename, NULL);
}

int CASSETTE_CreateRaw(char const *filename)
{
	return CreateBlankImage(A8CAS_FORMAT_RAW, filename, NULL);
}

int CASSETTE_FileSupported(char const *filename)
{
	A8CAS_info cas_info;
	A8CAS_FILE *file;

	cas_info.format = A8CAS_FORMAT_ANY;
	cas_info.mode = A8CAS_RDWR;
	if ((file = A8CAS_open(filename, &cas_info)) == NULL)
		return FALSE;
	A8CAS_close(file);
	return TRUE;
}

unsigned int CASSETTE_GetPosition(void)
{
	if (cassette_file == NULL)
		return 0;
	if (CASSETTE_position_in_blocks)
		return (unsigned int)A8CAS_tell(cassette_file) + 1;
	else
		return (unsigned int)(A8CAS_tell(cassette_file) / cassette_samplerate);
}

unsigned int CASSETTE_GetSize(void)
{
	if (cassette_file == NULL)
		return 0;
	if (CASSETTE_position_in_blocks)
		return (unsigned int)A8CAS_size(cassette_file);
	else
		return (unsigned int)(A8CAS_size(cassette_file) / cassette_samplerate);
}

void CASSETTE_Seek(unsigned int position)
{
	if (cassette_file != NULL) {
		if (CASSETTE_position_in_blocks) {
			if (position > 0)
				position --;
			A8CAS_seek(cassette_file, (unsigned long)position);
		} else
			A8CAS_seek(cassette_file, (unsigned long)position * cassette_samplerate);
		eof_of_tape = 0;
		signal_length = 0.0;
		CASSETTE_record = FALSE;
#ifdef SYNCHRONIZED_SOUND
		audio.samples_left_in_signal = 0;
#endif /* SYNCHORNIZED_SOUND */
		current_sig.rate = 1;
		UpdateFlags();
	}
}

/* set motor status
 1 - on, 0 - off
 remark: the real evaluation is done in AddScanLine*/
void CASSETTE_TapeMotor(int onoff)
{
	if (cassette_motor != onoff) {
		cassette_motor = onoff;
		UpdateFlags();
		ResetTurboState();
	}
}

int CASSETTE_ToggleWriteProtect(void)
{
	if (CASSETTE_status != CASSETTE_STATUS_READ_WRITE &&
	    CASSETTE_status != CASSETTE_STATUS_WRITE_ONLY)
		return FALSE;
	CASSETTE_write_protect = !CASSETTE_write_protect;
	UpdateFlags();
	return TRUE;
}

int CASSETTE_ToggleRecord(void)
{
	if (CASSETTE_status == CASSETTE_STATUS_NONE)
		return FALSE;
	CASSETTE_record = !CASSETTE_record;
	if (CASSETTE_record)
		eof_of_tape = 0;
	UpdateFlags();
	UpdateTurboState();
	/* Return FALSE to indicate that recording will not work. */
	return !CASSETTE_record || ((CASSETTE_status == CASSETTE_STATUS_READ_WRITE ||
	                             CASSETTE_status == CASSETTE_STATUS_WRITE_ONLY) && !CASSETTE_write_protect);
}

void CASSETTE_ToggleTurboState(void)
{
	CASSETTE_turbo_state = !CASSETTE_turbo_state;
	UpdateTurboState();
}

void CASSETTE_SetTurboType(enum CASSETTE_TURBO_TYPE_t type)
{
	CASSETTE_turbo_type = type;

	switch (type) {
	case CASSETTE_TURBO_AST_XC12:
		turbo_control = TURBO_CONTROL_COMMAND;
		turbo_descend_is_1 = 0;
		break;
	case CASSETTE_TURBO_BLIZZARD:
		turbo_control = TURBO_CONTROL_DATA_OUT;
		turbo_descend_is_1 = 0;
		break;
	case CASSETTE_TURBO_HARD:
		/* PWM circuit is enabled by asserting COMMAND and turning on the motor. */
		turbo_control = TURBO_CONTROL_COMMAND_MOTOR;
		turbo_descend_is_1 = 0;
		break;
	case CASSETTE_TURBO_MANUAL:
		turbo_control = TURBO_CONTROL_MANUAL;
		turbo_descend_is_1 = 0;
		break;
	case CASSETTE_TURBO_MANUAL_REV:
		turbo_control = TURBO_CONTROL_MANUAL;
		turbo_descend_is_1 = 1;
		break;
	default:
		turbo_control = TURBO_CONTROL_NONE;
	}
	CASSETTE_turbo_state = 0;
	ResetTurboState();
}

void CASSETTE_SetTurboTolerance(int value)
{
	CASSETTE_turbo_tolerance = value;
	if (cassette_file != NULL)
		UpdateTurboTolerance();
}

#ifdef SYNCHRONIZED_SOUND
void CASSETTE_SetCrosstalkVolume(unsigned int value)
{
	CASSETTE_crosstalk_volume = value;
	if (cassette_file != NULL)
		UpdateCrosstalkVolume();
}


/* 16 bit mixing */
static void mix(SWORD *dst, SWORD *src, int sndn, int volume)
{
	int s1;
	SWORD s2;
	int val;

	while (sndn--) {
		s1 = (int) *src;
		s1 = s1*volume/CASSETTE_DEFAULT_AUDIO_VOLUME;
		s2 = *dst;
		src++;
		val = s1 + s2;
		if (val > 32767) val = 32767;
		if (val < -32768) val = -32768;
		*dst++ = val;
		if (audio.n_pokeys == 2) {
			s2 = *dst;
			val = s1 + s2;
			if (val > 32767) val = 32767;
			if (val < -32768) val = -32768;
			*dst++ = val;
		}
	}
}

/* 8 bit mixing */
static void mix8(UBYTE *dst, UBYTE *src, int sndn, int volume)
{
	SWORD s1, s2;
	SWORD val;

	while (sndn--) {
		s1 = (SWORD)(*src) - 0x80;
		s1 = s1*volume/CASSETTE_DEFAULT_AUDIO_VOLUME;
		s2 = (SWORD)(*dst) - 0x80;
		src++;
		val = s1 + s2;
		if (val > 127) val = 127;
		if (val < -128) val = -128;
		*dst++ = (UBYTE)(val + 0x80);
		if (audio.n_pokeys == 2) {
			s2 = (SWORD)(*dst) - 0x80;
			val = s1 + s2;
			if (val > 127) val = 127;
			if (val < -128) val = -128;
			*dst++ = (UBYTE)(val + 0x80);
		}
	}
}

static void WriteSilence(int num_samples)
{
	if (audio.bit16)
		memset(audio.buffer + (audio.write_pos << 1), 0, num_samples << 1);
	else
		memset(audio.buffer + audio.write_pos, 127, num_samples);
	audio.write_pos += num_samples;
}

static void CassetteWriteAudio(int num_ticks)
{
	int num_samples = 0;

	signal_length -= num_ticks;
	if (signal_length < 0.0) {
		num_samples = ceil((-signal_length) / SYNCSOUND_ticks_per_sample);
		signal_length += num_samples * SYNCSOUND_ticks_per_sample;
	}
	if ((audio.bit16 ? (audio.write_pos + num_samples) << 1 : audio.write_pos + num_samples) > audio.buf_size) {
/*		Log_print("Cassette audio buffer overload: %f", signal_length);*/
		audio.write_pos = 0;
	}

/*	Log_print("caswrite: num_samples=%i", num_samples);*/
	if (cassette_writable && num_samples > 0) {
/*	Log_print("addscanline: numsamples=%i, signal=%u, tick=%i", num_samples, POKEY_serial_data_output, ANTIC_CPU_CLOCK);*/
		current_sig.signal = POKEY_serial_data_output;
		current_sig.length = num_samples;
		current_sig.rate = audio.samplerate;
		if (A8CAS_write_audio(cassette_file, &current_sig, audio.buffer + (audio.bit16 ? audio.write_pos << 1 : audio.write_pos)) != 0) {
			WriteSilence(num_samples);
			return;
		}
		audio.write_pos += num_samples;
	} else {
		WriteSilence(num_samples);
	}
}

static void CassetteReadAudio(int num_ticks)
{
	double const ticks_per_sec = (Atari800_tv_mode == Atari800_TV_NTSC ? Atari800_FPS_NTSC : Atari800_FPS_PAL) * 114 * Atari800_tv_mode;
	int num_samples = 0;

	if (cassette_readable) {
		signal_length -= num_ticks;
		while (signal_length < 0.0) {
			if (audio.samples_left_in_signal > 0) {
				int free_space = (int)((audio.bit16 ? audio.buf_size >> 1 : audio.buf_size) - audio.write_pos);
				unsigned int samples_to_read;
				if (free_space == 0) {
					audio.write_pos = 0;
#ifdef DEBUG_AUDIO					
					Log_print("Audio buffer overflow by %u samples", audio.samples_left_in_signal);
#endif /* DEBUG_AUDIO */
					continue;
				}
				samples_to_read = (free_space < audio.samples_left_in_signal ? free_space : audio.samples_left_in_signal);
				A8CAS_read_audio(cassette_file, audio.buffer + (audio.bit16 ? audio.write_pos << 1 : audio.write_pos), samples_to_read);
				audio.samples_left_in_signal -= samples_to_read;
				audio.sample_pos += samples_to_read * SYNCSOUND_ticks_per_sample;
				audio.write_pos += samples_to_read;
				continue;
			}

			if (A8CAS_read_with_audio(cassette_file, &current_sig, &audio.samples_left_in_signal) != 0) {
				/* Ignore the eventual error, assume it is the end of file */
				eof_of_tape = 1;
				UpdateFlags();
				break;
			}
			signal_length += (double)current_sig.length * ticks_per_sec / current_sig.rate;
		}
	}

	audio.sample_pos -= num_ticks;
	if (audio.sample_pos < 0.0) {
		num_samples = ceil((-audio.sample_pos) / SYNCSOUND_ticks_per_sample);
		audio.sample_pos += num_samples * SYNCSOUND_ticks_per_sample;
	}

	{
		int free_space = (int)((audio.bit16 ? audio.buf_size >> 1 : audio.buf_size) - audio.write_pos);
		if (free_space < num_samples) {
#ifdef DEBUG_AUDIO					
			Log_print("Audio buffer overflow by %u samples", num_samples);
#endif /* DEBUG_AUDIO */
			audio.write_pos = 0;
		}
	}

	if (cassette_readable) {
		int free_space = (int)((audio.bit16 ? audio.buf_size >> 1 : audio.buf_size) - audio.write_pos);
		unsigned int samples_to_read = (free_space > audio.samples_left_in_signal ? audio.samples_left_in_signal : free_space);
#ifdef DEBUG_AUDIO					
		if (samples_to_read < num_samples)
			Log_print("Not enough samples to read: %i", num_samples - samples_to_read);
#endif /* DEBUG_AUDIO */
		A8CAS_read_audio(cassette_file, audio.buffer + (audio.bit16 ? audio.write_pos << 1 : audio.write_pos), samples_to_read);
		audio.write_pos += samples_to_read;
		audio.samples_left_in_signal -= samples_to_read;
		num_samples -= samples_to_read;
		if (num_samples < 0)
			audio.sample_pos -= num_samples * SYNCSOUND_ticks_per_sample;
	}

	if (num_samples > 0)
		WriteSilence(num_samples);
}

void CASSETTE_InitAudio(unsigned int samplerate, unsigned int n_pokeys, int b16)
{
	audio.samplerate = samplerate;
	audio.n_pokeys = n_pokeys;
	audio.bit16 = b16;
	if (audio.buffer != NULL) {
		free(audio.buffer);
	}
	/* AUDIO.BUF_SIZE has to be a few (5 is OK) bytes larger than needed,
	   because number of CPU cycles per frame is not constant - it varies
	   a bit because CPU emulation stops only after performing a full 6502
	   instruction (which can last 2 to 6 cycles). */
	audio.buf_size = ((unsigned int)ceil(SYNCSOUND_samples_per_frame) + 5) * (b16 ? 2 : 1);
	audio.buffer = Util_malloc(audio.buf_size);
	
	audio.write_pos = 0;
	audio.sample_pos = 0.0;
	if (cassette_file != NULL)
		A8CAS_set_audio(cassette_file, samplerate, (b16 ? 16 : 8));
	cassette_write_func = &CassetteWriteAudio;
	cassette_read_func = &CassetteReadAudio;
}

void CASSETTE_UpdateSound(void *sndbuffer, int sndn)
{
	unsigned int numsamples = audio.write_pos;
	if (ESC_enable_sio_patch) {
		/* When SIO patch is active, CASSETTE_AddScanLine is not
		   called. Therefore we must update PREV_CPU_CLOCK here, so
		   that CASSETTE_AddScanLine won't fail horribly when the SIO
		   patch is turned off. */
		prev_cpu_clock = ANTIC_CPU_CLOCK;
		return;
	}
	if (!CASSETTE_play_audio || CASSETTE_audio_volume == 0)
		return;
	if (audio.n_pokeys == 2)
		sndn >>= 1;

	if (numsamples >= sndn) {
/*		Log_print("Update reserve: %i, sndn=%i, %i", numsamples - sndn, sndn, audio.buf_size);*/
		numsamples = sndn;
	}
/*	else {
		Log_print("Update underflow: %i", sndn - numsamples);
	}*/
	if (audio.bit16)
		mix((SWORD *)sndbuffer, (SWORD *)audio.buffer, numsamples, CASSETTE_audio_volume);
	else
		mix8((UBYTE *)sndbuffer, (UBYTE *)audio.buffer, numsamples, CASSETTE_audio_volume);
	if (numsamples >= sndn) {
		/* Audio buffer's size (numsamples) is computed in CASSETTE_InitAudio()
		   according to SYNCSOUND_samples_per_frame, so it is at most a few bytes
		   larger that sndn. Therefore the copying below would copy only a few
		   bytes. It's easier and simpler to do that instead of maintaining a
		   circular buffer. */
		if (audio.bit16)
			memmove(audio.buffer, audio.buffer + (sndn << 1), (audio.write_pos - sndn) << 1);
		else
			memmove(audio.buffer, audio.buffer + sndn, audio.write_pos - sndn);
		audio.write_pos -= sndn;
	} else {
		audio.write_pos = 0;
	}
}
#endif /* SYNCHRONIZED_SOUND */

static void CassetteWrite(int num_ticks)
{
/*	Log_print("caswrite: num_samples=%i", num_samples);*/
	if (cassette_writable) {
/*	Log_print("addscanline: numsamples=%i, signal=%u, tick=%i", num_samples, POKEY_serial_data_output, ANTIC_CPU_CLOCK);*/
		current_sig.signal = POKEY_serial_data_output;
		current_sig.length = num_ticks;
		current_sig.rate = (Atari800_tv_mode == Atari800_TV_NTSC ? Atari800_FPS_NTSC : Atari800_FPS_PAL) * 114 * Atari800_tv_mode;
		if (A8CAS_write(cassette_file, &current_sig) != 0) {
			return;
		}
	}
}

static void CassetteRead(int num_ticks)
{
	if (cassette_readable) {
		double const ticks_per_sec = (Atari800_tv_mode == Atari800_TV_NTSC ? Atari800_FPS_NTSC : Atari800_FPS_PAL) * 114 * Atari800_tv_mode;
		signal_length -= num_ticks;
		while (signal_length < 0.0) {
			if (A8CAS_read(cassette_file, &current_sig) != 0) {
				/* Ignore the eventual error, assume it is the end of file */
				eof_of_tape = 1;
				UpdateFlags();
				return;
			}
			signal_length += (double)current_sig.length * ticks_per_sec / current_sig.rate;
/*			Log_print("signal_length: %f", signal_length);*/
		}
	}
}

void CASSETTE_AddScanLine(void)
{
	int num_ticks = (int)(ANTIC_CPU_CLOCK - prev_cpu_clock);
	prev_cpu_clock = ANTIC_CPU_CLOCK;
	ResetTurboState();

	if (CASSETTE_record)
		(*cassette_write_func)(num_ticks);
	else
		(*cassette_read_func)(num_ticks);
}

/* Returns status of I/O-line */
int CASSETTE_IOLineStatus(void)
{
	/* if motor off and EOF return always 1 (equivalent the mark tone) */
	if (cassette_motor == 0
	    || eof_of_tape != 0
	    || cassette_file == NULL) {
		return 1;
	}
	if (!ESC_enable_sio_patch)
		CASSETTE_AddScanLine();
	if (CASSETTE_turbo_state && turbo_descend_is_1)
		return !(int)current_sig.signal;
	else
		return (int)current_sig.signal;
}

/* --- Functions for loading/saving with SIO patch --- */

int CASSETTE_AddGap(int gaptime)
{
	gap_delay += gaptime;
	if (gap_delay < 0)
		gap_delay = 0;
	return gap_delay;
}

/* Indicates that a loading leader is expected by the OS */
void CASSETTE_LeaderLoad(void)
{
	if (CASSETTE_record)
		CASSETTE_ToggleRecord();
	CASSETTE_TapeMotor(TRUE);
	gap_delay = 9600;
	/* registers for SETVBV: third system timer, ~0.1 sec */
	CPU_regA = 3;
	CPU_regX = 0;
	CPU_regY = 5;
}

/* indicates that a save leader is written by the OS */
void CASSETTE_LeaderSave(void)
{
	if (!CASSETTE_record)
		CASSETTE_ToggleRecord();
	CASSETTE_TapeMotor(TRUE);
	gap_delay = 19200;
	/* registers for SETVBV: third system timer, ~0.1 sec */
	CPU_regA = 3;
	CPU_regX = 0;
	CPU_regY = 5;
}

/* Reads a cassette record with SIO Patch active.
   LENGTH is a wanted length of a record (must be > 0).
   DEST_ADDR is an Atari memory address to store the bytes to.
   Returns 1 on load without errors, and 0 when the record ended too early, or
   checksum was wrong, or there was a file read error. */
int CASSETTE_ReadToMemory(UWORD dest_addr, int length)
{
/*	Log_print("ReadWithPatch! length: %u, dest_addr: %u, gap_delay: %i, signal_length: %f", length, dest_addr, gap_delay, signal_length);*/
	CASSETTE_TapeMotor(1);
	if (!cassette_readable)
		return 0;
	{
		double delay_s = (double) gap_delay / 1000.0; /* delay in seconds, in floating-point */
		if (signal_length > 0.0) {
			double const ticks_per_sec = (Atari800_tv_mode == Atari800_TV_NTSC ? Atari800_FPS_NTSC : Atari800_FPS_PAL) * 114 * Atari800_tv_mode;
			delay_s -= signal_length / ticks_per_sec;
			signal_length = 0.0;
		}

			
		while (delay_s > 0.0) {
			/* Skip signals. */
			if (A8CAS_read(cassette_file, &current_sig) != 0) {
				/* Ignore the eventual error, assume it is the end of file */
				gap_delay = delay_s * 1000.0;
				eof_of_tape = 1;
				UpdateFlags();
				return 0;
			}
			delay_s -= (double)current_sig.length / current_sig.rate;
/*			Log_print("Skipped; delay: %f", delay_s);*/
		}
		gap_delay = 0;
	}

	/* Skip any remaining non-byte signals */
	{
		A8CAS_signal buffer[256];
		while (A8CAS_read_signals(cassette_file, buffer, 256) != 0);
		if (A8CAS_error(cassette_file) != A8CAS_ERR_NONE) {
			eof_of_tape = 1;
			UpdateFlags();
			return 0;
		}
	}

	/* Load bytes */
	{
		UBYTE buffer[256]; /* Atari records aren't typically longer than 132 bytes anyway. */
		unsigned int checksum = 0;
		length ++; /* Increase length so that checksum will be also read. */

		for (;;) {
			int i;
			unsigned int baudrate; /* Ignored. */
			unsigned int fill = A8CAS_read_bytes(cassette_file, buffer, (length > 256 ? 256: length), &baudrate);

			if (fill == 0) /* Record ended too early, or read error. */
				break;
/*			for (i = 0; i < fill; i ++)
				fprintf(stderr, " %02x", buffer[i]);*/
			length -= fill;
			if (length == 0)
				/* End of a record - so don't treat the last byte (checksum) as data.
				   After this, the read checksum is found at buffer[fill]. */
				fill --;
			MEMORY_CopyToMem(buffer, dest_addr, fill); /* Checkum is not copied. */
			/* Update checksum. */
			for (i = 0; i < fill; i ++)
				checksum += buffer[i];
			if (length == 0) { /* Wanted number of bytes has been read */
				while (checksum > 0xff)
					checksum = (checksum & 0xff) + (checksum >> 8);
/*				fprintf(stderr, " fill: %u, dest_addr: %u, check: %02x, good? %i\n", fill, dest_addr, checksum, checksum == buffer[fill]);*/
				return checksum == buffer[fill];
			}
			dest_addr += fill;
		}
		/* Deal with an error. */
		if (A8CAS_error(cassette_file) != A8CAS_ERR_NONE) { /* Real read error */
			eof_of_tape = 1;
			UpdateFlags();
		}
	}
	/* If we reached here, there was either a read error or the record ended too early. */
	return 0;
}

/* Writes a cassette record with SIO Patch active.
   LENGTH is the record's length (must be > 0).
   SRC_ADDR is an Atari memory address where the record starts.
   Returns 1 on load without errors, and 0 when there was a file write error. */
int CASSETTE_WriteFromMemory(UWORD src_addr, int length)
{
/*	Log_print("WriteWithPatch! length: %u, src_addr: %u, gap_delay: %i, signal_length: %f", length, src_addr, gap_delay, signal_length);*/
	CASSETTE_TapeMotor(1);
	if (!cassette_writable)
		return 0;

	/* Write leader. */
	if (gap_delay > 0) {
		current_sig.signal = 1;
		current_sig.length = gap_delay;
		current_sig.rate = 1000;
		if (A8CAS_write(cassette_file, &current_sig) != 0)
			return 0;
		gap_delay = 0;
	}

	/* Write block bytes. */
	{
		UBYTE buffer[256];
		unsigned int checksum = 0;

		for (;;) {
			int i;
			unsigned int to_write = (length > 255 ? 255 : length);
		
			length -= to_write;
			MEMORY_CopyFromMem(src_addr, buffer, to_write);
			/* Update checksum. */
			for (i = 0; i < to_write; i ++)
				checksum += buffer[i];
			if (length == 0) {
				/* End of a record - also write checksum. */
				while (checksum > 0xff)
					checksum = (checksum & 0xff) + (checksum >> 8);
				buffer[to_write ++] = checksum;
				return !A8CAS_write_bytes(cassette_file, buffer, to_write, 600);
			}
			src_addr += to_write;
		}
	}
}


/*
vim:ts=4:sw=4:
*/
